/// Josh's Logger
/// A Redacted Software Product
/// Developed & Maintained by Josh O'Leary
/// This work is dedicated to the public domain.

#pragma once

#include <string>
#include <Event.h>
#include <jlog/color_codes.hpp>

#ifdef __linux__
    #define FUNCTION __PRETTY_FUNCTION__
#endif

#ifdef _WIN32
    #define FUNCTION __FUNCSIG__
#endif

namespace jlog
{
    /// Severity levels for logging. Higher numbers filter less messages out.
    /// @see LOG_LEVEL macro
    enum class severity : uint8_t
    {
        none,       // Show no output
        fatal,      // Show only fatal errors
        error,      // Show fatal and not-necessarily-fatal errors
        warning,    // Show warnings additionally
        verbose,    // Show errors, warnings, and 'verbose' output (i.e. extra information)
        debug,      // Show debugging messages (Basically everything).
    };

    inline severity loglevel = severity::debug; // Default log level always debug

    struct LogEntry
    {
        severity level;
        std::string content;
        std::string timestamp;
    };

    // TODO: Fully implement logging callback.
    static Event<LogEntry> on_log = Event<LogEntry>();

    inline std::vector<LogEntry> log_history;

    /// A single piece of a logging message, with color code, content, and delimiter
    /// These are strung together to build full logger messages in a flexible manner.
    struct token
    {
        colorcode colorCode = color_codes::FG_DEFAULT;
        std::string content = "";
        std::string delimiter = "[]";
    };

    void set_default_logfile(const std::string& filename);

    std::string get_timestamp();

    /// Writes an input string directly to standard output
    /// @note Does not append a newline to the message.
    void log_to_console(const std::string& message);
    /// Writes an input string directly to the default destination logfile
    /// @note Does not append a newline to the message.
    /// @see set_default_logfile()
    void log_to_file(const std::string& message);
    /// Writes an input string directly to the specified destination logfile
    /// @note Does not append a newline to the message.
    void log_to_file(const std::string& filename, const std::string& message);
    /// Writes an input string directly to the specified output stream
    /// @note Does not append a newline to the message.
    void log_to_stream(std::ostream stream, const std::string& message);

    /// Parses a sequence of tokens into a printable message, and logs to the following destinations:
    /// standard output (console)
    /// output file (default)
    /// logging callback (Event)
    /// @note This file is implemented differently per-platform to handle differences in console color handling.
    /// @see windows/jlog.cpp linux/jlog.cpp
    void log(std::vector<token> tokens);

    // Generic formatters for building loggers.

    std::vector<token> trace_format(
        const std::string& func,
        const std::string& file,
        int line);

    std::vector<token> log_format(
        const std::string& severity_name,
        const std::string& message,
        const colorcode& severity_cc = color_codes::FG_WHITE);

    std::vector<token> log_detailed_format(
    const std::string& severity_name,
    const std::string& message,
    const std::string& func,
    const std::string& file,
    int line,
    const colorcode& severity_cc = color_codes::FG_WHITE);

    // Predefined generic loggers for jlog.
    std::vector<token> info_format(const std::string& message);

    std::vector<token> info_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    std::vector<token> warning_format(const std::string& message);

    std::vector<token> warning_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    std::vector<token> error_format(const std::string& message);

    std::vector<token> error_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    std::vector<token> fatal_format(const std::string& message);

    std::vector<token> fatal_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    std::vector<token> verbose_format(const std::string& message);

    std::vector<token> verbose_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);

    std::vector<token> debug_format(const std::string& message);

    std::vector<token> debug_detailed_format(
        const std::string &message,
        const std::string &func,
        const std::string &file,
        int line);
}

#define INFO(i)    if (jlog::loglevel >= jlog::severity::none) { jlog::log(jlog::info_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define VERBOSE(i) if (jlog::loglevel >= jlog::severity::verbose) { jlog::log(jlog::verbose_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define DEBUG(i)   if (jlog::loglevel >= jlog::severity::debug) { jlog::log(jlog::debug_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define WARNING(i) if (jlog::loglevel >= jlog::severity::warning) { jlog::log(jlog::warning_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define ERROR(i)   if (jlog::loglevel >= jlog::severity::error) { jlog::log(jlog::error_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define FATAL(i)   if (jlog::loglevel >= jlog::severity::fatal) { jlog::log(jlog::fatal_detailed_format(i, FUNCTION, __FILE__, __LINE__)); }

#define LOGLEVEL(i) jlog::loglevel = i;

